
/*
    This file is part of the MoreQt library - A try to make Qt more cute
    Copyright (C) 2001, 2002  Yves Bailly, kafka@kafka-fr.net

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <math.h>

#include <mq_dbl_range_control.h>
#define __MQ_DBL_RANGE_CONTROL_CPP__ 1
#include "mq_dbl_range_control_p.h"

/*!
    \class mqDblRangeControl
    \brief Provides a QRangeControl-like class that acts on \a double rather on \a int.
    
    The range is defined by a minimum and a maximum values. Because \a double values are
    (almost) continuous, it is needed to define a granularity. This is achived either
    by giving a number of steps between the two extremas, or by giving a step size.
    For example, if you set a range between -10.0 and 10.0 :
    - giving a number of steps of 100 means that you can get values -10.0, -9.8, ...
      and so on until 9.8, 10.0 : the step size is (10.0 - (-10.0)) / 100 = 0.2 ;
    - giving a step size of 0.1 results in a number of steps of 200.
    
    \todo
    The meaning of \c lineStep() and \c pageStep() is not yet very clear and should
    be defined.
*/

mqDblRangeControl::mqDblRangeControl()
{
    d = new mqDblRangeControlPrivate ;
    setNbSteps(1000) ;
}

mqDblRangeControl::mqDblRangeControl(double minValue, double maxValue, unsigned nb_steps,
                                     unsigned line_step, unsigned page_step, double value)
{
    d = new mqDblRangeControlPrivate ;
    d->mDblMinValue = minValue ;
    d->mDblMaxValue = maxValue ;
    d->mDblValue = bound(value) ;
    setNbSteps(nb_steps) ;
    d->mLineStep = line_step ;
    d->mPageStep = page_step ;
}

mqDblRangeControl::mqDblRangeControl(double minValue, double maxValue, double step_size,
                                     unsigned line_step, unsigned page_step, double value)
{
    d = new mqDblRangeControlPrivate ;
    d->mDblMinValue = minValue ;
    d->mDblMaxValue = maxValue ;
    d->mDblValue = bound(value) ;
    setStepSize(step_size) ;
    d->mLineStep = line_step ;
    d->mPageStep = page_step ;
}

mqDblRangeControl::mqDblRangeControl(const QRangeControl& rc, unsigned nb_steps)
{
    d = new mqDblRangeControlPrivate() ;
    d->mDblMinValue = rc.minValue() ;
    d->mDblMaxValue = rc.maxValue() ;
    d->mDblValue = bound(rc.value()) ;
    setNbSteps(nb_steps) ;
    d->mLineStep = rc.lineStep() ;
    d->mPageStep = rc.pageStep() ;
}

mqDblRangeControl::mqDblRangeControl(const QRangeControl& rc, double step_size)
{
    d = new mqDblRangeControlPrivate ;
    d->mDblMinValue = rc.minValue() ;
    d->mDblMaxValue = rc.maxValue() ;
    d->mDblValue = bound(rc.value()) ;
    setStepSize(step_size) ;
    d->mLineStep = rc.lineStep() ;
    d->mPageStep = rc.pageStep() ;
}

// virtual
mqDblRangeControl::~mqDblRangeControl()
{ 
    delete d ;
    d = 0 ;
}

void mqDblRangeControl::setupQRangeControl(QRangeControl& rc) const
{
    rc.setRange(0, d->mNbSteps) ;
    synchQRangeControl(rc) ;
}

void mqDblRangeControl::synchFromQRangeControl (const QRangeControl& rc)
{
    setValue(mapIntValue(rc.value())) ;
}

void mqDblRangeControl::synchQRangeControl (QRangeControl& rc) const
{
    rc.setValue(mapDblValue(d->mDblValue)) ;
}

double mqDblRangeControl::dblValue() const
{ return d->mDblValue ; }

void mqDblRangeControl::setValue(double val)
{ d->mDblValue = bound(val) ; }

double mqDblRangeControl::dblMinValue () const
{ return d->mDblMinValue ; }

double mqDblRangeControl::dblMaxValue () const
{ return d->mDblMaxValue ; }

void mqDblRangeControl::setRange (double minValue, double maxValue)
{
    d->mDblMinValue = minValue < maxValue ? minValue : maxValue ;
    d->mDblMaxValue = minValue < maxValue ? maxValue : minValue ;
    if ( d->mDblValue < d->mDblMinValue )
        setValue(d->mDblMinValue) ;
    else
        if ( d->mDblMaxValue < d->mDblValue )
            setValue(d->mDblMaxValue) ;
    UpdateSteps() ;
}

double mqDblRangeControl::dblRange() const
{ return d->mDblMaxValue - d->mDblMinValue ; }

void mqDblRangeControl::setMinValue (double minVal)
{
    if ( d->mDblMaxValue < minVal )
        return ;
    d->mDblMinValue = minVal ;
    UpdateSteps() ;
}

void mqDblRangeControl::setMaxValue (double maxVal)
{
    if ( maxVal < d->mDblMinValue )
        return ;
    d->mDblMaxValue = maxVal ;
    UpdateSteps() ;
}

double mqDblRangeControl::bound (double v) const
{
    if ( v > d->mDblMaxValue )
        return d->mDblMaxValue ;
    else
        if ( v < d->mDblMinValue )
            return d->mDblMinValue ;
        else
            return v ;
}

double mqDblRangeControl::stepSize() const
{ return d->mStepSize ; }

unsigned mqDblRangeControl::nbSteps() const
{ return d->mNbSteps ; }

void mqDblRangeControl::setStepSize(double step_size)
{
    d->mPreferStepSize = true ;
    if ( step_size == d->mStepSize )
        return ;
    d->mStepSize = step_size ;
    UpdateSteps() ;
}

void mqDblRangeControl::setNbSteps(unsigned nb_steps)
{
    d->mPreferStepSize = false ;
    if ( nb_steps == d->mNbSteps )
        return ;
    d->mNbSteps = nb_steps ;
    UpdateSteps() ;
}

double mqDblRangeControl::mapIntValue(unsigned value) const
{ return bound(d->mDblMinValue + value*d->mStepSize) ; }

unsigned mqDblRangeControl::mapDblValue(double value) const
{ return (unsigned)(floor((bound(value) - d->mDblMinValue) / d->mStepSize)) ; }

unsigned mqDblRangeControl::stepDblValue() const
{ return mapDblValue(d->mDblValue) ; }

// private
void mqDblRangeControl::UpdateSteps()
{
    if ( d->mPreferStepSize )
    {
        if ( d->mDblMinValue < d->mDblMaxValue )
            d->mNbSteps = (unsigned)
                (ceil((d->mDblMaxValue - d->mDblMinValue)/d->mStepSize)) ;
        else
            d->mNbSteps = 1 ;
    }
    else
    {
        if ( d->mDblMinValue < d->mDblMaxValue )
            d->mStepSize = (d->mDblMaxValue - d->mDblMinValue)/d->mNbSteps ;
        else
            d->mStepSize = 0.1 ;
    }
}

